module Main where

import Data.List
import Text.Parsec

main :: IO ()
main = do
  raw <- getContents
  putStrLn $ "day5a: " ++ show (solveA $ validate $ map (parse (lineP createSpots) []) $ lines raw)
  putStrLn $ "day5b: " ++ show (solveB $ validate $ map (parse (lineP createSpots') []) $ lines raw)

validate :: [Either ParseError [(Int,Int)]] -> [(Int,Int)]
validate [] = []
validate ((Left _):_) = error "invalid input"
validate ((Right b):bs) = b ++ (validate bs)

lineP :: ((Int,Int) -> (Int,Int) -> [(Int,Int)]) -> Parsec String () [(Int,Int)]
lineP f = do
  x1 <- many1 digit
  _ <- char ','
  y1 <- many1 digit
  _ <- string " -> "
  x2 <- many1 digit
  _ <- char ','
  y2 <- many1 digit
  return $ f (read x1, read y1) (read x2, read y2)

range :: Int -> Int -> [Int]
range x y = [ i .. j ]
  where
    (i,j) = if x > y then (y,x) else (x,y)

createSpots :: (Int,Int) -> (Int,Int) -> [(Int,Int)]
createSpots (x1,y1) (x2,y2)
  | x1 == x2 = [ (x1,j) | j <- range y1 y2]
  | y1 == y2 = [ (i,y1) | i <- range x1 x2]
  | otherwise = []

createSpots' :: (Int,Int) -> (Int,Int) -> [(Int,Int)]
createSpots' (x1,y1) (x2,y2)
  | x1 == x2 = [ (x1,j) | j <- range y1 y2]
  | y1 == y2 = [ (i,y1) | i <- range x1 x2]
  | otherwise = runDiag (x1,y1) (x2,y2)

runDiag :: (Int,Int) -> (Int,Int) -> [(Int,Int)]
runDiag (x1,y1) (x2,y2)
  | y1 < y2 =
    if x1 < x2
    then take (x2-x1+1) $ iterate dl (x1,y1)
    else take (x1-x2+1) $ iterate dr (x1,y1)
  | otherwise =
    if x1 < x2
    then take (x2-x1+1) $ iterate dr (x2,y2)
    else take (x1-x2+1) $ iterate ur (x1,y1)
  where
    dl (x,y) = (x+1,y+1)
    dr (x,y) = (x-1,y+1)
    ur (x,y) = (x-1,y-1)

rle :: Eq a => [a] -> [(Int,a)]
rle xs = rle' $ zip (repeat 1) xs

rle' :: Eq a => [(Int,a)] -> [(Int,a)]
rle' ((xv,xk):(yv,yk):xs) =
  if xk == yk
  then rle' $ ((xv+yv),xk) : xs
  else (xv,xk) : (rle' ((yv,yk):xs))
rle' xs = xs

solveA :: [(Int,Int)] -> Int
solveA ps = length $ filter (\x -> (fst x) >= 2) $ rle $ sort ps

solveB :: [(Int,Int)] -> Int
solveB ps = length $ filter (\x -> (fst x) >= 2) $ rle $ sort ps
